/**
 * @brief 708 reader class implementation.
 * @file read708.cpp
 * @version 1.5.1
 * @date March 4, 2016
 * @details Class for reading ARINC 708 data from a file. The base class
 * implements the ARINC 708 standard.
 * 
 * Version History:
 * Version 0.0 on December 5, 2013
 * Version 0.1 on December 19, 2013
 * Version 0.2 on December 23, 2013
 * Version 0.2.1 on December 23, 2013
 * Version 0.3 on January 6, 2014
 * Version 1.0 on January 13, 2014
 * Version 1.1 on March 6, 2014
 * Version 1.2 on March 18, 2014
 * Version 1.3 on May 7, 2014
 * Version 1.5.1 on March 4, 2016
 */

#include <iostream>
#include <math.h>
#include <stdio.h>
#include "read708.h"

using namespace reader708;

read708::read708(string fileName) throw (exception) {
    nDataWords = 96; // = 512 range bins * 3 bits each / (2 bytes per word * 8 bits per bytes)
   
    try {
        this->initialize(fileName);
    } catch (exception &e) {
        throw;
    }  
}

read708::read708(string fileName, unsigned int numDataWords) throw (exception) {
    nDataWords = numDataWords;
    
    try {
        this->initialize(fileName);
    } catch (exception &e) {
        throw;
    }    
}

read708::~read708() {
    fclose(in_flnme);
    delete[] buffer;
    delete[] blockData;
}

void read708::initialize(string fileName) throw(exception) {
    if ((in_flnme = fopen(fileName.c_str(),"rb")) == NULL) {
        throw runtime_error("Error opening 708 file for input.");
    }
    
    try {
        buffer = new unsigned short[HEADERWORDS+nDataWords];
    } catch (bad_alloc &ba) {
        cerr << "bad_alloc caught: " << ba.what() << endl;
        throw runtime_error("Cannot allocate memory for data buffer.");
    } 

    try {
        blockData = new unsigned short[nDataWords];
    } catch (bad_alloc &ba) {
        cerr << "bad_alloc caught: " << ba.what() << endl;
        throw runtime_error("Cannot allocate memory for block data.");
    }
    this->resetValues();
}

void read708::resetValues() {
    azimuth = 0.0f;
    elevation = 0.0f;
    range = 0.0f;
    gain = 0.0f;
    for (int i = 0; i < NRANGBINS; i++) {
        displayValue[i] = 0;
    }
}
   
void read708::readFrame() throw (exception) {
    unsigned int i;
    size_t nread;

    nread = fread(buffer,2,nDataWords+HEADERWORDS,in_flnme); 
    if (nread != nDataWords+HEADERWORDS) {
         throw endOfFileException();
    }
    if ((buffer[0] == 948)) {   
        for (i = 0; i < HEADERWORDS; i++)
          headerData[i] = buffer[i];
        for (i = 0; i < nDataWords; i++)
          blockData[i] = buffer[i+HEADERWORDS];
        this->decodeHeader();
        this->decodeFrame();
    } else {
        this->resetValues();
        cerr << "Data error in frame." << endl;
        throw runtime_error("Data error in frame.");
    }
}  /* end of read708() */

void read708::decodeHeader() {
    int temp;
    
    // decode azimuth
    temp = (int)headerData[3];
    azimuth = (float)(((temp&0x7ff8)<<17)>>20) * 0.087890626f;
    //protect the system from a runaway programmer.
    if (azimuth > 90.0f) {
        azimuth =  89.8f;
        cerr << "Warning: Azimuth angle out of bounds." << endl;
    }
    if (azimuth <-90.0f) {
        azimuth = -89.8f;
        cerr << "Warning: Azimuth angle out of bounds." << endl;
    }
 
    // decode elevation
    elevation = (float)((headerData[2]&0x000f)<<3);
    elevation = (elevation +(float)((headerData[1]>>13)&0x0007) )*0.25f;
    if (((headerData[2]>>3)&0x0001) == 1) {elevation = elevation - 32;}
    if (elevation > 90.0f) {
        elevation =  89.8f;
        cerr << "Warning: Elevation angle out of bounds." << endl;
    }
    if (elevation <-90.0f) {
        elevation = -89.8f;
        cerr << "Warning: Elevation angle out of bounds." << endl;
    }
            
    // decode range
    temp = (headerData[2] >> 10); // decode range setting
    // ARINC 708 standard range is provided in nautical miles
    range = (temp == 0) ? 320.0f : (float)(temp*5);

    // decode gain
    temp = (int)(((headerData[2]>>10)<<6));
    temp = (headerData[2]>>4)-temp;
    this->determineGain(temp);
}

void read708::determineGain(int gainWord) {
    switch(gainWord) {
        case 0:
            gain = 0.0f; // max
            break;
        case 7:
            gain = -5.0f;
            break;
        case 11:
            gain = -11.0f;
            break;
        case 62:
            gain = -62.0f;
            break;
        case 63:
            gain = 0.0f; // cal
            break;
        default:
            cerr << "Unknown gain setting. Gain set to zero" << endl;
            gain = 0.0f;
            break;
    }
}

void read708::decodeFrame() {
    unsigned int atmp,btmp;
    unsigned int bin, word, i;

    // Valid 708 data records begin with label 0xb4 (octal 55).
    // step through the data words stored in packed format --
    // each range bin contains 3-bit values
    word = 0;
    bin = 0;
    while (word < nDataWords) { // value of size of array bf453  
        atmp = ((blockData[word+1]<<16)+blockData[word]);
        word++;
        for (i=0; i<16; i++) {
            if (i == 7) {
                word++;
                btmp = (unsigned int)blockData[word];
                atmp = (btmp<<11)+atmp;
            }
            displayValue[bin] = (unsigned int)(atmp&0x0007);
            bin++;
            if (fmod(bin,NRANGBINS)==0) {break;} // make sure does not go above array bounds
            atmp = atmp>>3;             // get 3 bit values
	}	
        word++;
    }
}

float read708::getAzimuth() {
    return azimuth;
}

float read708::getElevation() {
    return elevation;
}

float read708::getGain() {
    return gain;
}

float read708::getRange() {
    return range;
}

unsigned int read708::getDisplayValue(unsigned int bin) {
    unsigned int value;
    if (bin >= NRANGBINS) {
        value = 0;
        cerr << "Warning: Range bin must be 0-511. Zero returned." << endl;
    } else {
        value = displayValue[bin];
    }
    return value;
}